/*
 * Roger Burton West <roger@firedrake.demon.co.uk>
 */

#include "ratfile.h"
#include <assert.h>

#ifndef SYSEOL
#define SYSEOL "\r\n"
#endif

#define RET_STAT(x) status=x; return status

// public functions

RATFile::RATFile(const char *fullfilename,const char *mode)
{ // constructor with params, open or fail quietly
  status=RAT_CLOSED;
  status=Open(fullfilename,mode);
}

RATFile::RATFile()
{ // base constructor
  status=RAT_CLOSED;
}

int RATFile::Open(const char *fullfilename,const char *cmode)
{ // file open routine: open, create index
  XString mode,line;
  fpos_t ivalue;
  bool stat_inline,stat_lastchar,stat_eol,stat_bl;
  char load,eol=0;
  bool newf;

  filename=fullfilename;
  searchdir=0;
  writeable=false;
  append=false;
  tempfile=false;
  livepointcount=0;
  if (status == RAT_OK) { // something odd's going on
    RET_STAT(RAT_ERR);
  }
  mode=cmode;
  mode.ToLower();
  if (mode.Find('a') != -1) { append=true; writeable=true; }
  if (mode.Find('w') != -1) { writeable=true; append=false; }
  if (mode.Find('t') != -1) { tempfile=true; writeable=true; append=false; }
  if (!tempfile) { // check special modes
    if (mode.Find('f') != -1) { // "fail if file exists"
      filehandle=fopen((const char *)filename,"rb");
      if (filehandle) { // fail
        fclose(filehandle);
        RET_STAT(RAT_ERR);
      }
    }
    if (mode.Find('n') != -1) { // "delete file if it exists"
      filehandle=fopen((const char *)filename,"rb");
      if (filehandle) { // delete file
        fclose(filehandle);
        remove((const char *)filename);
      }
    }
  }
  newf=true;
  if (!tempfile) { // check for file's prior existence
    filehandle=fopen((const char *)filename,"rb");
    if (filehandle) { // file existed before
      fclose(filehandle);
      newf=false;
    } else {
      if (mode.Find('r') != -1 && writeable==false) { // file should have existed
        RET_STAT(RAT_ERR);
      }
    }
  }
  if (tempfile) { // special opening routine
    filehandle=tmpfile();
  } else { // standard open
    if (newf) {
      filehandle=fopen((const char *)filename,"wb+");
    } else {
      filehandle=fopen((const char *)filename,"rb");
    }
  }
  if (!filehandle) { // can't open file
    RET_STAT(RAT_ERR);
  }
  indexhandle=tmpfile();
  if (mode=="r") {
    workhandle=filehandle;
  } else {
    workhandle=tmpfile();
  }
  if ((!indexhandle) || (!workhandle)) { // can't create index/temp
    fclose(workhandle);
    fclose(indexhandle);
    fclose(filehandle);
    RET_STAT(RAT_ERR);
  }
  if (!newf) { // file existed before
    stat_lastchar=false;
    stat_inline=true;
    stat_eol=false;
    end=0;
    line="";
    do { // for each line in file
      fgetpos(workhandle,&ivalue);
      fread(&load,1,1,filehandle);
      if (!feof(filehandle)) { // except the non-existent one
        stat_bl=false;
        if (load=='\r' || load=='\n') { // EOL of some sort
          stat_inline=false;
          if (!stat_eol) { // new EOL
            eol=load;
          } else { // old (persisting) EOL
            if ((stat_eol) && eol==load) { // blank line!
              stat_bl=true;
              line="";
            }
          }
          stat_eol=true;
        } else { // not EOL
          stat_eol=false;
        }
        if (load >= ' ') stat_inline=true;
        if (stat_inline) { // add to temp string
          if (stat_eol) { // EOL, so don't bother
            line="";
          } else { // not EOL
            line|=load;
          }
        }
        if (((!stat_lastchar) && (stat_inline)) || stat_bl) { // reached start of line
          if (fwrite(&ivalue,sizeof(ivalue),1,indexhandle) != 1) { // failed to extend
            fclose(filehandle);
            fclose(indexhandle);
            fclose(workhandle);
            RET_STAT(RAT_ERR);
          }
          end++;
        }
        if (((stat_lastchar) && (!stat_inline)) || stat_bl) { // reached end of line
          line |= SYSEOL;
          if (mode!="r") {
            fwrite((const char *)line,line.Len(),1,workhandle);
          }
          line="";
        }
        stat_lastchar=stat_inline;
      }
    } while (!feof(filehandle));
    if (stat_inline) { // final line
      line |= SYSEOL;
      fwrite((const char *)line,line.Len(),1,workhandle);
      line="";
    }
  } else { // file didn't exist before
    end=0;
  }
  fflush(indexhandle);
  fflush(filehandle);
  fflush(workhandle);

  ratpr=0;
  if (append) { // all writes at end
    ratpw=end;
  } else { // all writes at read pointer
    ratpw=0;
  }
  RET_STAT(RAT_OK);
}

RATPOS RATFile::GetPos()
{ // return read pointer
  if (status != RAT_OK) { // unless something's up
    return (RATPOS)(-1);
  }
  return ratpr;
}

RATPOS RATFile::SetPos(RATPOS posinput,int mode)
{ // set read (and write?) pointer
  RATPOS position;
  if (status != RAT_OK) { // unless something's up
    status=RAT_ERR;
    return 0;
  }
  switch (mode) { // generate real position
    case RAT_CUR:
      position=ratpr+posinput;
      break;
    case RAT_END:
      position=end+posinput;
      break;
    default:
      position=posinput;
  }
  if (position<0) position=0;
  if (position>end) position=end;
  ratpr=position;
  if ((writeable) && (!append)) { // not append-to-end
    ratpw=ratpr;
  }
  return ratpr;
}

RATPOS RATFile::GetLength()
{
  return end;
}

RATPOS *RATFile::GetLivePos()
{ // get a "live" position, i.e. changes with Insert/Delete
  if (status != RAT_OK) { // unless something's up
    return (RATPOS *)NULL;
  }
  posit[livepointcount]=ratpr;
  return &posit[livepointcount++];
}

RATPOS RATFile::SetLivePos(RATPOS *livepos)
{ // use a "live" position
  if (!livepos) return ratpr; // null pointer
  if (*livepos < 0) return ratpr; // invalid pointer (deleted line)
  return SetPos(*livepos);
}

XString RATFile::Read()
{ // get a line
  fpos_t ivalue;
  char buffer;
  XString output;
  int pos;

  if (status != RAT_OK) { // unless something's up
    RET_STAT(RAT_ERR);
  }
  if (ratpr >= end) { // past EOF
    ratpr=end+1;
    return (XString)"";
  }
  fseek(indexhandle,sizeof(ivalue)*ratpr,SEEK_SET);
  fread(&ivalue,sizeof(ivalue),1,indexhandle);
  fflush(indexhandle);
  fseek(workhandle,ivalue,SEEK_SET);
  pos=-1;
  do { // for each char in line
    pos++;
    fread(&buffer,1,1,workhandle);
    if (buffer >= ' ') output |= buffer;
  } while (buffer != '\n' && buffer != '\r');
  ratpr++;
  if (!append) ratpw=ratpr;
  fflush(workhandle);
  return output;
}

int RATFile::Write(const char *datum)
{ // write a line
  char eol[]=SYSEOL;
  fpos_t ivalue;

  if (status != RAT_OK) { // unless something's up
    RET_STAT(RAT_ERR);
  }
  if (!writeable) { // or it's not open for writing
    RET_STAT(RAT_ERR);
  }

  fseek(workhandle,(fpos_t)0,SEEK_END);
  fgetpos(workhandle,&ivalue);
  fwrite(datum,strlen(datum),1,workhandle);
  fwrite(eol,strlen(eol),1,workhandle);
  if (ratpw == end) { // really append a line
    end++;
  }
  fseek(indexhandle,sizeof(ivalue)*ratpw,SEEK_SET);
  fwrite(&ivalue,sizeof(ivalue),1,indexhandle);
  fflush(indexhandle);
  ratpw++;
  if (!append) ratpr=ratpw;
  fflush(workhandle);
  RET_STAT(RAT_OK);
}

int RATFile::InsertLine(int lines)
{ // insert blank lines BEFORE current line, fpointer on first blank
  XString sstr;
  fpos_t ivalue;
  FILE *tempfilehandle;
  int count;
  char eol[]=SYSEOL;

  if (status != RAT_OK) { // unless something's up
    RET_STAT(RAT_ERR);
  }
  if ((!writeable) || append) { // or it's not open for random writing
    RET_STAT(RAT_ERR);
  }

  if (lines < 1) { //pointless!
    return(RAT_OK);
  }

  if (ratpw > end) { // don't need to insert, just append lines
    for(count=0;count<lines;count++) { // each line
      Write("");
    }
    RET_STAT(RAT_OK);
  }

  tempfilehandle=tmpfile();

  fseek(indexhandle,(fpos_t)0,SEEK_SET);
  Copy(indexhandle,tempfilehandle,ratpw*sizeof(fpos_t));

  fgetpos(workhandle,&ivalue);
  fwrite(eol,strlen(eol),1,workhandle);
  for (count=0;count<lines;count++) { // each line
    fwrite(&ivalue,sizeof(ivalue),1,tempfilehandle);
  }

  fseek(indexhandle,sizeof(ivalue)*ratpw,SEEK_SET);
  Copy(indexhandle,tempfilehandle,(end-ratpw)*sizeof(fpos_t));

  if (livepointcount) { // have to update live pointers
    unsigned long c;
    for (c=0;c<livepointcount;c++) {
      if (posit[c] >= ratpr) {
        posit[c] += lines;
      }
    }
  }

  end += lines;

  fclose(indexhandle);
  indexhandle=tempfilehandle;

  fflush(indexhandle);
  fflush(workhandle);

  RET_STAT(RAT_OK);
}

int RATFile::DeleteLine(int lines)
{ // delete lines from file
  fpos_t ivalue;
  FILE *tempfilehandle;

  if (status != RAT_OK) { // unless something's up
    RET_STAT(RAT_ERR);
  }
  if ((!writeable) || append) { // or it's not open for random writing
    RET_STAT(RAT_ERR);
  }

  if (lines < 1) { //pointless!
    return(RAT_OK);
  }

  if (ratpw >= end) { // irrelevant
    RET_STAT(RAT_OK);
  }

  if (lines+ratpw > end) { // too many deletions
    lines=end-ratpw;
  }

  tempfilehandle=tmpfile();

  fseek(indexhandle,(fpos_t)0,SEEK_SET);
  Copy(indexhandle,tempfilehandle,ratpw*sizeof(fpos_t));

  fseek(indexhandle,sizeof(ivalue)*(ratpw+lines),SEEK_SET);
  Copy(indexhandle,tempfilehandle,(end-(ratpw+lines))*sizeof(fpos_t));

  if (livepointcount) { // have to update live pointers
    unsigned long c;
    for (c=0;c<livepointcount;c++) {
      if (posit[c] >= ratpw) {
        if (posit[c] >= ratpw+lines) {
          posit[c] -= lines;
        } else {
          posit[c] = -1;
        }
      }
    }
  }

  end -= lines;

  fclose(indexhandle);
  indexhandle=tempfilehandle;

  fflush(indexhandle);
  fflush(workhandle);

  RET_STAT(RAT_OK);
}

XString RATFile::FindR(regex_t *datum)
{
  searchdir=1;
  memcpy(&regholder,datum,sizeof(regex_t));
  return FindNext();
}

XString RATFile::Find(const char *datum)
{
  regex_t tempreg;
  searchdir=1;
  regcomp(&tempreg,datum,REG_NOSPEC | REG_NEWLINE | REG_NOSUB);
  return FindR(&tempreg);
}

XString RATFile::FindBackR(regex_t *datum)
{
  searchdir=-1;
  memcpy(&regholder,datum,sizeof(regex_t));
  return FindNext();
}

XString RATFile::FindBack(const char *datum)
{
  regex_t tempreg;
  searchdir=-1;
  regcomp(&tempreg,datum,REG_NOSPEC | REG_NEWLINE | REG_NOSUB);
  return FindBackR(&tempreg);
}

XString RATFile::FindNext()
{
  XString match;
  char *cstr;
  int result;

  if (searchdir==0) return (XString)"";
  do {
    match=Read();
    if (searchdir==-1) {
      if (ratpr<2) {
        ratpr=end+1;
      } else SetPos(-2L,RAT_CUR);
    }
    cstr=(char *)malloc(match.Len()+1);
    strcpy(cstr,(const char *)match);
    result=regexec(&regholder,cstr,0,NULL,0);
    free(cstr);
  } while ((result!=0) && (ratpr<=end));
  if (result==0) {
    return match;
  } else return (XString)"";
}

int RATFile::Close()
{ // rationalise file, delete index
  if (status != RAT_OK) { // unless something's up
    RET_STAT(RAT_ERR);
  }
  /*
  if (livepointcount) { // nuke the live pointers
    RATLIVEPOS *current,*del;
    current=latest;
    for (;livepointcount>0;livepointcount--) {
      assert(current != NULL);
      del=current;
      current=current->next;
      free(del);
    }
  }
  */
  if ((writeable) && (!tempfile)) { // rationalise the file
    XString buffer;

    fclose(filehandle);
    filehandle=fopen((const char *)filename,"wb");
    ratpr=0;
    do { // loop over all lines
      buffer=Read();
      if (!Eof()) { // output line
        buffer |= SYSEOL;
        fwrite((const char *)buffer,buffer.Len(),1,filehandle);
      }
    } while (!Eof());
    fclose(filehandle);
    fclose(workhandle);
  } else { // just close
    fclose(filehandle);
  }
  fclose(indexhandle);
  regfree(&regholder);
  RET_STAT(RAT_CLOSED);
}

int RATFile::Eof()
{ // return end-of-file status
  if (ratpr > end) { // i.e. past, not at, end
    return true;
  }
  return false;
}

RATFile::~RATFile()
{ // destroy variable
  if (status == RAT_OK) { // closing if necessary
    Close();
  }
}

// private functions

int RATFile::Copy(FILE *srchandle, FILE *desthandle, fpos_t bytes)
{ // assumes positions already set!
  fpos_t tocopy,count;
  char *buffer;

  if (bytes == 0) return RAT_OK;

  fflush(srchandle);
  fflush(desthandle);

  tocopy=bytes;
  count=bytes;
  if (count > RATBUFSIZE) count=RATBUFSIZE;

  buffer=(char *)malloc(count);
  if (!buffer) { // after all my hard work, too...
    RET_STAT(RAT_ERR);
  }
  do {
    fread(buffer,1,count,srchandle);
    fwrite(buffer,1,count,desthandle);
    tocopy -= count;
    if (tocopy < RATBUFSIZE) count=RATBUFSIZE;
  } while (tocopy > 0);
  free(buffer);

  fflush(srchandle);
  fflush(desthandle);
  return RAT_OK;
}
