
/* Player toolkit */

/*
 * Based on:
 *
 * A.M. van den Bos' PlanMap header files
 * Micheal Keane's ALTSPEC.H
 *
 * Modification history:
 *
 * CC	97.11.13	Made Cargo structure to hold cargo
 * 			for ships, planets etc.
 *
 * CC	97.11.17	Removed old cargo fields replaced by Cargo.
 *
 * RBW 98.1.26 Incorporated CC's changes to live version
 */

#ifndef _tkplayer_h
#define _tkplayer_h

#include <limits.h>
#include <xstring.h>


#if INT_MIN < -2147483647 && INT_MAX >= 2147483647

typedef int            SB4;
typedef unsigned int   UB4;

#else

typedef long           SB4;
typedef unsigned long  UB4

#endif

#if SHRT_MIN < -32767 && SHRT_MAX >= 32767

typedef short          SB2;
typedef unsigned short UB2;

#else

typedef int            SB2;
typedef unsigned int   UB2

#endif

typedef signed char    SB1;
typedef unsigned char  UB1;

#define NUM_PLANETS        500   // PLANETS has a fixed number of planets
#define PLANETNAME_LENGTH  20    // As in file PLANET.NM
#define NUM_RACES          11    // There are 11 races
#define MAXSHIP            500   // Maximum number of ships
#define MAXHULL            120   // Maximum number of hulls that can be read
#define MAXBEAM            10    // Beam specs
#define MAXTORP            10    // Torp specs
#define MAXENG             9     // Engine specs
#define MAXMESSAGESIZE  (41*15)  // 15 lines of 40 characters + CR
#define MAXMESSAGE         50    // Max. number of outgoing messages

/* Length of race names (long form, plural form, adjective form) */
#define RACE_LENGTH(j) (((j)==0)?30:(((j)==1)?20:12))

#define NO_HULL_STR "CANNOT SCAN HULL              "

/**********************************************************************/
/* Maximum number of structures on a planet as function of the number */
/* of clans. (function-like macro; there must be at least 1 clan)     */
/**********************************************************************/

#define MAX_MINES(clans) ( ((clans)<=200) ? (clans) : 200+(UB2)(0.5+sqrt((clans)-200)))
#define MAX_FACT(clans)  ( ((clans)<=100) ? (clans) : 100+(UB2)(0.5+sqrt((clans)-100)))
#define MAX_DEF(clans)   ( ((clans)<= 50) ? (clans) :  50+(UB2)(0.5+sqrt((clans)- 50)))

/**********************************************************************/
/* Crew kill/damage/shields vs Beam/torp kill and mass                */
/**********************************************************************/

#define SQR(x) ((x)*(x))  // rise to the power of two
#define CREW_KILL(kill,mass)     (min(9999L,(UB4)((kill)*(80.0/((mass)+1)))))
#define SHIELD_DAMAGE(expl,mass) (min(100L,(UB4)((expl)*(80.0/((mass)+1))+1)))
#define SHIP_DAMAGE(expl,mass)   (min(100L,(UB4)((expl)*SQR(80.0/((mass)+1)))+2))

/**************************************************************************/
/* Colonists/natives status (build-screen VGA Planets)                    */
/**************************************************************************/

#define COL_STATUS(clans,tax,planet,mines,factories)        \
((int) (1000.0-sqrt((float)clans) -80*(tax) -               \
3*abs((int)50-(planet)) -((float)(mines)+(factories))/3))

#define NAV_STATUS(clans,gov,tax,planet,mines,factories)    \
((int) (1000.0-sqrt((float)clans) -85*(tax) -               \
50*((int)10-(gov)) -((float)(mines)+(factories))/2))

#define TEXT_STATUS(status)                      \
((status >=  450) ? "Love you":                  \
 (status >=   50) ? "Like your leadership":      \
 (status >= -100) ? "Are undecided about you":   \
 (status >= -550) ? "Are angry about you":       \
                    "Hate you!" )

#define SHORT_STATUS(status)     \
((status >=  450) ? "Love":      \
 (status >=   50) ? "Like":      \
 (status >= -100) ? "Neutral":   \
 (status >= -550) ? "Angry":     \
                    "Hate!")

/**************************************************************************/
/* Mineral and cargo amounts                                              */
/**************************************************************************/

struct Minerals
{
  enum Type
  {
    FUEL,
    MONEY,
    TRI,
    DUR,
    MOL,
    SUPPLIES,
    CLANS,
    CARGO_ITEMS
  };

  UB4  tri;
  UB4  dur;
  UB4  mol;
  static UB4 nothing;

  virtual UB4 &operator[](size_t type)
  {
    switch (type)
    {
      case TRI:         return tri;
      case DUR:         return dur;
      case MOL:         return mol;
      default:          nothing = 0;
                        return nothing;
    }
  }

  virtual UB4 Mass()
  {
    return tri + dur + mol;
  }
};

struct BuildAmount : public Minerals
{
  UB4  money;

  virtual UB4 &operator[](size_t type)
  {
    switch (type)
    {
      case MONEY:       return money;
      case TRI:         return tri;
      case DUR:         return dur;
      case MOL:         return mol;
      default:          nothing = 0;
                        return nothing;
    }
  }

  virtual UB4 Mass()
  {
    return tri + dur + mol;
  }
};

struct Resources : public Minerals
{
  UB4  fuel;

  virtual UB4 &operator[](size_t type)
  {
    switch (type)
    {
      case FUEL:        return fuel;
      case TRI:         return tri;
      case DUR:         return dur;
      case MOL:         return mol;
      default:          nothing = 0;
                        return nothing;
    }
  }

  virtual UB4 Mass()
  {
    return tri + dur + mol + fuel;
  }
};

struct Cargo : public Minerals
{
  UB4  fuel;
  UB4  money;
  UB4  supplies;
  UB4  clans;

  virtual UB4 &operator[](size_t type)
  {
    switch (type)
    {
      case FUEL:        return fuel;
      case MONEY:       return money;
      case TRI:         return tri;
      case DUR:         return dur;
      case MOL:         return mol;
      case SUPPLIES:    return supplies;
      case CLANS:       return clans;
      default:          nothing = 0;
                        return nothing;
    }
  }

  virtual UB4 Mass()
  {
    return tri + dur + mol + supplies + clans;
  }
};


/**************************************************************************/
/* Coordinates                                                            */
/**************************************************************************/

struct Coord
{
  UB2  x;
  UB2  y;

  Coord() {}

  Coord(UB2 xx, UB2 yy)
  {
    x = xx;
    y = yy;
  }

  Coord& operator += (const Coord &coord)
  {
    x += coord.x;
    y += coord.y;
    return *this;
  }

  Coord& operator -= (const Coord &coord)
  {
    x -= coord.x;
    y -= coord.y;
    return *this;
  }

  Coord operator+ (const Coord &coord)
  {
    return Coord(x+coord.x, y+coord.y);
  }

  Coord operator- (const Coord &coord)
  {
    return Coord(x-coord.x, y-coord.y);
  }

  UB4 operator* (const Coord &coord) // square of distance
  {
    Coord temp;
    temp=*this-coord;
    return temp.x*temp.x+temp.y*temp.y;
  }

  bool operator== (const Coord &coord)
  {
    return (coord.x == x && coord.y == y);
  }

  bool operator!= (const Coord &coord)
  {
    return (coord.x != x || coord.y != y);
  }

};



/* NOTE! ALL ARRAYS ARE 1-BASED! */

/**************************************************************************/
/* RACE.NM file                                                           */
/**************************************************************************/

typedef struct
{ XString sFull;
  XString sPlural;
  XString sAdjective;
} RACENM;

/**************************************************************************/
/* GEN??.DAT file with player's scores, turn number and checksums         */
/**************************************************************************/

/* Note: Corrected by RPGBW */

typedef struct    // Score of a player
{ UB2  wPlanets;
  UB2  wCapitalShips;
  UB2  wFreighters;
  UB2  wBases;
} SCORE;

typedef struct
{ XString sDateHost;      // ASCII string [10]
  XString sTimeHost;       // ASCII string [8]
  SCORE Scores[NUM_RACES+1];// Score of each player (player's nr. = index)
  UB2   wPlayer;
  XString sPassword;        // (10)
  XString sSignature1;      // (10)
  XString sSignature2;      // (10)
  UB4   lCS_Ship;           // UB1 -wise checksum over SHIP??.DAT  & SHIP??.DIS  (offset 0x81)
  UB4   lCS_Planet;         // UB1 -wise checksum over PDATA??.DAT & PDATA??.DIS (offset 0x85)
  UB4   lCS_Base;           // UB1 -wise checksum over BDATA??.DAT & BDATA??.DIS (offset 0x89)
  UB2   wIsNewPwd;          /* Is there a new password? */
  XString sNewPass;       /* if so, here is the new password (10)
                               (rot50 encryption) */
  UB2   wTurnNr;            /* Current turn number           (offset 0x99) */
  UB2   wHostID;            /* Timestamp checksum */
} GENDATA;

#define TURN_OFFSET      0x99  // Offset in GEN??.DAT for turn number

/**************************************************************************/
/* XYPLAN.DAT file with (x,y) coordinates of the 500 planets.             */
/* The file is 3000 bytes and can be loaded directly into a               */
/* PLANET_COORDS[NUM_PLANETS] stucture. The dummy field is not used and   */
/* always (?) zero. The 1-based index is the ID of the planet.            */
/*                                                                        */
/* PLANET.NM file with the 500 planet names. 10000 bytes in size and can  */
/* be loaded directly into a CHAR[NUM_PLANETS][20] structure. A 1-based   */
/* index is the ID of the planet.                                         */
/**************************************************************************/

typedef struct
{  Coord pos;
   UB2  owner;
} PLANET_COORDS;

typedef struct
{  XString sName; // [20]
} PLANET_NAME;

/***************************************************************************/
/* Hull, engine, beam & torpedoes specifications.                          */
/* VGA Planets calculates a single UB4  checksum over the specification    */
/* files HULLSPEC.DAT, ENGSPEC.DAT, BEAMSPEC.DAT & TORPSPEC.DAT by summing */
/* the BYTES of each file.                                                 */
/***************************************************************************/

typedef struct
{ UB2 wHullnum;
} TRUEHULL;

typedef struct              // HULLSPEC.DAT
{ XString sName;            // Name of the hull [30]
  UB2  wPicture;            // Picture ID
  UB2  wDamagePic;          // Always 1, perhaps part of Picture ID; I don't know
  BuildAmount cargo;        // Amount of resources to build
  UB2  wFuel, wCrew;        // Size of fuel tanks & crew
  UB2  wEngines, wMass;     // Engines needs & Mass
  UB2  wTechLevel, wCargo;  // Tech level needs and max cargo
  UB2  wFighterBays;        // Fighter bays
  UB2  wTorpsBays;          // Torp bays
  UB2  wBeams;              // Max. number of beams
} HULLSPEC;

typedef struct              // BEAMSPEC.DAT
{
  XString sName;            // Name of beam weapon [20]

  BuildAmount cargo;        // Amount of resources to build

  UB2  wMass;               // Total mass of beam weapon
  UB2  wTechLevel;          // Tech level required to build
  UB2  wKill;               // Kill level of weapon
  UB2  wExplosive;          // Destructive level of weapon
} BEAMSPEC;

typedef struct              // TORPSPEC.DAT
{
  XString sName;            // Name of torp. launcher [20]

  BuildAmount cargo;        // Amount of resources to build

  UB2  wCostTorp;           // Costs of one torpedo of this type
  UB2  wMass;               // Total mass of torp. launcher
  UB2  wTechLevel;          // Tech level required to build torp. launcher
  UB2  wKill;               // Kill level of weapon
  UB2  wExplosive;          // Destructive level of weapon
} TORPSPEC;

typedef struct              // ENGSPEC.DAT
{
  XString sName;            // Name of engine [20]

  BuildAmount cargo;        // Amount of resources to build

  UB2  wTechLevel;          // Tech level required to build it.
  UB4  lFuelUse[10];        // Warp x requires lFuelUse[x]
} ENGSPEC;

/***************************************************************************/
/* SHIP??.DAT First UB2  of file #ships, followed by a list of ships.      */
/* SHIP.HST   First UB2  is a dummy, followed by an ARRAY[1..500] OF       */
/*            SHIPDATA. The index is the ship's ID                         */
/***************************************************************************/

typedef struct
{ UB2  wID;                 // ship's ID
  UB2  wRace;               // owner of ship (should be = playernum)
  XString sCode;            // Friendly code; three characters [3]
  UB2  wWarp;               // Speed of ship
  Coord pos;                // Location of ship (x, y)
  Coord dis;                // Displacement next turn (tan(bearing) = x/y)
  UB2  wDriveLevel;         // Tech level of engines
  UB2  wShipType;           // Type of ship (index in HULLSPEC.DAT)
  UB2  wBeamLevel;          // Tech level of beams
  UB2  wBeamWeapons;        // Number of beam weapons
  UB2  wFighterBays;        // Number of fighter bays
  UB2  wTorpLevel;          // Torp. tech. level
  UB2  wTorps;              // Number of torps/fighters
  UB2  wTorpsBays;          // Number of torp. lauchers
  UB2  wMission;            // Mission of the ship
  UB2  wEnemy;              // Primairy enemy
  UB2  wTowed;              // Ship being towed
  UB2  wDamage;             // Ships damage in %
  UB2  wCrew;               // Number of crew members on ship
  XString  sName;           // Name of ship [20]
  Cargo cargo;              // combined cargo structure
  Cargo toplanet;
  UB2 wPID;
  Cargo toship;
  UB2 wSID;

  UB2  wIntercept;          // ID of ship to intercept
  HULLSPEC hullspec;
  ENGSPEC engspec;
  BEAMSPEC beamspec;
  TORPSPEC torpspec;
} SHIPDATA;

/**************************************************************************/
/* TARGET??.DAT file with scanned enemy hulls.                            */
/**************************************************************************/

typedef struct
{ UB2  wID;                // ID of ship
  UB2  wRace;              // Owner
  UB2  wWarp;              // Current speed
  Coord pos;               // Location
  UB2  wHull;              // Index in HULLSPEC.DAT
  UB2  wBearing;           // In degrees
  XString sName;           // Name of ship [20]
} TARGETDATA;

/***********************************************************************/
/* SHIPXY??.DAT & SHIPXY.HST. File with (scanned) enemy ships          */
/* 500 entries, position in file is ship's ID                          */
/***********************************************************************/

typedef struct
{ Coord pos ;               // Location of ship
  UB2  wOwner;              // wOwner = 0 -> no ship scanned with this ID
  UB2  wMass;               // Mass of ship
} SHIPXY;

/*************************************************************************/
/* PDATA??.DAT & PDATA.HST first word in file equals #planets or 0 resp. */
/*************************************************************************/

typedef struct
{ Coord pos;                // brought in from elsewhere
  XString sName;            // brought in from elsewhere
  UB2  wOwner;              // Owner (0 = no owner)
  UB2  wID;                 // ID of planet
  XString sCode;            // Friendly code [3]
  UB2  wMines;              // Number of mineral mines
  UB2  wFactories;          // Number of factories
  UB2  wDefense;            // Defences on planet

  Cargo cargo;              // combined cargo structure
  Resources left;           // minerals left in planet
  Resources density;        // density of minerals

  UB2  wColTax;             // Colonist-taxes in %
  UB2  wNativeTax;          // Native-taxes in %
  SB2  iColStatus;          // <20 fighting, 20 rioting, 40 very angry, 50 unhappy, 70 calm, 90 happy
  SB2  iNativeStatus;       // id.

  UB2  wGov;                // Government of natives (6=monarcy 5=feudel etc.)
  UB4  lNatives;            // Number of natives
  UB2  wRace;               // Native's race (7=amph 8=ghip etc.)

  UB2  wTemperature;        // Climate. Defines planet picture & color-palette for picture (?)
  UB2  BuildNewBase;        // 1 if building
} PLANETDATA;

/**************************************************************************/
/* MESS??.DAT file with outgoing messages.                               */
/* First word is #messages in file                                       */
/* Then an ARRAY[50] OF MESSAGE_ENTRY (see below)                        */
/* Then the actual messages, add an offset (+13) to each letter          */
/* Note: no more than 50 messages can be transmitted                     */
/*************************************************************************/

typedef struct
{ UB4  lOffset;             // Start+1 in file
  UB2  wLength;             // Length of message in bytes
  UB2  iFrom;               // Sender of message
  UB2  iTo;                 // Destination of message
  XString sMessage;
} OUTMESS;

/*************************************************************************/
/* MDATA??.DAT is the file with incoming messages.                       */
/* First word is #messages in file                                       */
/* Then an ARRAY[#messages] OF { UB4  lOffset; UB2  wLength}             */
/* Then the actual messages, add an offset of -13 to each letter         */
/*************************************************************************/

typedef struct
{ UB4  lOffset;             // Start+1 in file
  UB2  wLength;             // Length of message in bytes
  XString sMessage;
} INMESS;

/*************************************************************************/
/* VCR??.DAT. File for the Visual Combat Recorder.                       */
/* The first UB2  of the file is #battles.                               */
/* ( Thanks Alistair )                                                   */
/*************************************************************************/

/* Note: Not tested */

typedef struct
{
   UB2  wMass;              // Mass of ships
   XString sName;           // Name of ship [20]
   UB2  wDamage;            // Initial damage state of ship
   UB2  Crew;               // Initial crew on ship
   UB2  wID;                // ID of ship
   UB2  wRace;              // Owner of ship
   UB2  wBitmap;            // Index in RESOURCE.PLN
   UB2  wBeamTech;          // Tech. level of beam weapons
   UB2  wBeams;             // Number of beam weapons
   UB2  wBays;              // Number of fighter bays
   UB2  wTorpTech;          // Tech. level of torp. launchers
   UB2  wTorpFighters;      // Initial number of torps. or fighter on ship
   UB2  wTorps;             // Number of torp. lauchers
   UB2  wShield;            // Initial shields of both ships.
} COMBATANT;

typedef struct
{  UB4  RandomSeed;
   UB2  wPlanetTemp;        /* Temperature of planet */
   UB2  IsPlanet;           /* Tells whether right side is a planet or base */
   COMBATANT combatant[3];  // combatants
} VCR_DATA;

/**************************************************************************/
/* BDATA??.DAT  First UB2  #bases, followed by a list of bases.           */
/**************************************************************************/

typedef struct
{ UB2  wID;
  UB2  wOwner;
  UB2  wDefence;       // Defence of starbase
  UB2  wDamage;        // Damage in %

  /* Tech levels */
  UB2  wTechEngine;
  UB2  wTechHull;
  UB2  wTechBeam;
  UB2  wTechTorp;

  /* Items on starbase */
  UB2  wEngines[10];    // #engines on starbase index 0=stardrive 1    0x12
  UB2  wHulls[21];      // #Hulls, index = index in truehull           0x24
  UB2  wBeams[11];      // #beam weapons on starbase index 0=Laser     0x4C
  UB2  wTLaunchers[11]; // #torp lauchers on base, 0=Mark 1            0x60
  UB2  wTorps[11];      // #tops on starbase index 0=Mark 1            0x74

  UB2  wFighters;       // #fighters on base 0x88
  UB2  wFixID;          // ID of ship being fixed or recycled
  UB2  wFixShip;        // 0=do nothing, 1=fix ship, 2=recycle ship
  UB2  wMission;        // 1 = Refuel, 2 = Max defence, ...

  /* Ship being build */
  UB2  wShip;           // Ship being build (index in Hullspel?)
  UB2  wEngineLevel;    // Engine level of ship being build
  UB2  wBeamLevel;      // Beam level of ship being build
  UB2  wBeamOnShip;     // Number of beams on ship
  UB2  wTorpLevel;      // Torp. laucher tech level
  UB2  wTorpOnShip;     // Number of torp. laucher on ship

  UB2  wFighterOnShip;  // ignore, equals 0

} BASEDATA;

typedef struct
{ XString reginfo[3];   // [25]
} FIZZBIN;

typedef struct
{  XString sName; // [20]
} ION_NAME;

extern BASEDATA *read_all_basedata(bool datfile=true);
extern BASEDATA *read_basedata(int snum, bool datfile=true);
extern BASEDATA *read_basedata_A(int bnum, bool datfile=true);
extern BEAMSPEC *read_all_beamspec();
extern BEAMSPEC *read_beamspec(int bnum);
extern bool del_inmess(int mnum);
extern bool del_outmess(int mnum);
extern bool set_game_dir(const char *gdir);
extern bool set_player(int player);
extern bool write_basedata(int snum, BASEDATA *str);
extern bool write_basedata_A(int bnum, BASEDATA *str);
extern bool write_outmess(int mnum, OUTMESS *str);
extern bool write_planetdata(int snum, PLANETDATA *str);
extern bool write_planetdata_A(int pnum, PLANETDATA *str);
extern bool write_shipdata(int snum, SHIPDATA *str);
extern bool write_shipdata_A(int snum, SHIPDATA *str);
extern ENGSPEC *read_all_engspec();
extern ENGSPEC *read_engspec(int e_num);
extern FILE *mkopen(const char *filename, bool trylocal, char mode);
extern GENDATA *read_gendata();
extern HULLSPEC *read_all_hullspec();
extern HULLSPEC *read_hullspec(int hnum);
extern INMESS *read_all_inmess();
extern INMESS *read_inmess(int mnum);
extern int num_bases(bool datfile=true);
extern int num_inmess();
extern int num_outmess();
extern int num_planets(bool datfile=true);
extern int num_ships(bool datfile=true);
extern int num_targets();
extern int num_vcrs();
extern ION_NAME *read_all_ion_name();
extern ION_NAME *read_ion_name(int inum);
extern OUTMESS *read_all_outmess();
extern OUTMESS *read_outmess(int mnum);
extern PLANETDATA *read_all_planetdata(bool datfile=true);
extern PLANETDATA *read_planetdata(int pnum, bool datfile=true);
extern PLANETDATA *read_planetdata_A(int pnum, bool datfile=true);
extern PLANET_COORDS *read_all_planet_coords();
extern PLANET_COORDS *read_planet_coords(int pnum);
extern PLANET_NAME *read_all_planet_name();
extern PLANET_NAME *read_planet_name(int pnum);
extern RACENM *read_all_racenm();
extern RACENM *read_racenm(int rnum);
extern SHIPDATA *read_all_shipdata(bool datfile=true);
extern SHIPDATA *read_shipdata(int snum, bool datfile=true);
extern SHIPDATA *read_shipdata_A(int snum, bool datfile=true);
extern SHIPXY *read_all_shipxy();
extern SHIPXY *read_shipxy(int snum);
extern TARGETDATA *read_all_targetdata();
extern TARGETDATA *read_targetdata(int tnum);
extern TORPSPEC *read_all_torpspec();
extern TORPSPEC *read_torpspec(int tnum);
extern TRUEHULL *read_all_truehull();
extern TRUEHULL *read_all_truehull(int pn);
extern TRUEHULL *read_truehull(int rnum);
extern TRUEHULL *read_truehull(int rnum, int pn);
extern VCR_DATA *read_all_vcr_data();
extern VCR_DATA *read_vcr_data(int vnum);
extern void make_checksum(bool ships, bool planets, bool bases);
extern XString rcode();
extern bool extraplan;
extern bool extraship;
extern bool write_fizzbin(FIZZBIN *str);
extern FIZZBIN *read_fizzbin();

#endif

